from jaseci.jac.ir.passes.ir_pass import IrPass
from jaseci.jac.jsci_vm.disasm import DisAsm


class PrinterPass(IrPass):
    def __init__(self, *args):
        super().__init__(*args)

    def enter_node(self, node):
        print("Entering", node)
        if hasattr(node, "bytecode"):
            DisAsm().disassemble(node.bytecode)

    def exit_node(self, node):
        print("Exiting", node)



import io
import sys
from unittest import TestCase

import jaseci.jac.tests.dot_code as dtc
from jaseci.actor.sentinel import Sentinel
from jaseci.graph.graph import Graph
from jaseci.svc import MetaService
from jaseci.utils.utils import TestCaseHelper


class TestDot(TestCaseHelper, TestCase):
    """Unit test for DOT support"""

    def setUp(self):
        super().setUp()
        self.sent = Sentinel(m_id="anon", h=MetaService().build_hook())
        self.gph = Graph(m_id="anon", h=self.sent._h)

        self.old_stdout = sys.stdout
        self.new_stdout = io.StringIO()
        sys.stdout = self.new_stdout

    def tearDown(self):
        sys.stdout = self.old_stdout
        super().tearDown()

    def to_screen(self):
        sys.stdout = self.old_stdout
        print("output: ", self.new_stdout.getvalue())
        sys.stdout = self.new_stdout

    def test_dot_node(self):
        """Test node in dot"""
        self.sent.register_code(dtc.dot_node)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_dot_node_no_dot_id(self):
        """Test node in dot with no do graph name"""
        self.sent.register_code(dtc.dot_node_no_dot_id)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_spawn_graph_node(self):
        """Test node in using spawn graphs instead of dot"""
        self.sent.register_code(dtc.spawn_graph_node)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "graph_root_node_name\n")

    def test_dot_node_multi_stmts(self):
        """Test node in dot, defined with multiple statements."""
        self.sent.register_code(dtc.dot_node_multi_stmts)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "real_test_node\n" "2021\n")

    def test_dot_edge(self):
        """Test edge in dot."""
        self.sent.register_code(dtc.dot_edge)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "root\n" "node_1\n" "node_2\n")

    def test_dot_edge_with_attrs(self):
        """Test edge in dot with attrs"""
        self.sent.register_code(dtc.dot_edge_with_attrs)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "root\n" "node_1\n")

    def test_dot_edge_with_attrs_vars(self):
        """Test edge in dot with attrs and variables"""
        self.sent.register_code(dtc.dot_edge_with_attrs_vars)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        self.assertEqual(self.new_stdout.getvalue(), "edge_1\n" "node_1\n")

    def test_dot_graph_parses(self):
        self.sent.register_code(dtc.dot_graph)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()

    def test_dot_str(self):
        self.sent.register_code(dtc.dot_graph)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        dot_str = self.gph.graph_dot_str()
        self.assertTrue("strict digraph root" in dot_str)

    def test_dot_quoted_string(self):
        self.sent.register_code(dtc.dot_quoted_string)
        gen_walker = self.sent.run_architype("init")
        gen_walker.prime(self.gph)
        gen_walker.run()
        report = gen_walker.report
        self.assertEqual(report, ["root", "this has space", "another space"])



from jaseci.jac.jsci_vm.op_codes import JsOp, JsAttr, type_map
from jaseci.jac.jsci_vm.inst_ptr import InstPtr, from_bytes
from base64 import b64decode


class DisAsm(InstPtr):
    def __init__(self):
        InstPtr.__init__(self)
        self._asm = []

    def disassemble(self, bytecode, print_out=True):
        if type(bytecode) == str:
            bytecode = b64decode(bytecode.encode())
        self._bytecode = bytearray(bytecode)
        while self._ip < len(self._bytecode):
            op = JsOp(self._bytecode[self._ip])
            if hasattr(self, f"dis_{op.name}"):
                getattr(self, f"dis_{op.name}")()
            else:
                self._asm.append([op.name])
            self._ip += 1
        if print_out:
            self.print()

    def print(self):
        for i in self._asm:
            print(*i)

    def dis_LOAD_CONST(self):  # noqa
        typ = JsAttr(self.offset(1))
        operand2 = self.offset(2)
        val = None
        if typ in [JsAttr.TYPE]:
            self._asm.append([self.cur_op(), typ.name, JsAttr(operand2).name])
            self._ip += 2
        elif typ in [JsAttr.INT, JsAttr.STRING]:
            val = from_bytes(type_map[typ], self.offset(3, operand2))
            self._asm.append([self.cur_op(), typ.name, operand2, val])
            self._ip += 2 + operand2
        elif typ in [JsAttr.FLOAT]:
            val = from_bytes(float, self.offset(2, 8))
            self._asm.append([self.cur_op(), typ.name, val])
            self._ip += 2 + 8
        elif typ in [JsAttr.BOOL]:
            val = bool(self.offset(2))
            self._asm.append([self.cur_op(), typ.name, val])
            self._ip += 2 + 1

    def dis_DEBUG_INFO(self):  # noqa
        byte_len_l = self.offset(1)
        line = from_bytes(int, self.offset(2, byte_len_l))
        byte_len_f = self.offset(3)
        jacfile = from_bytes(str, self.offset(4, byte_len_f)) if byte_len_f else 0
        self._asm.append([self.cur_op(), byte_len_l, line, byte_len_f, jacfile])
        self._ip += 2 + byte_len_l + byte_len_f



from jaseci.jac.ir.passes import IrPass
from jaseci.jac.jsci_vm.op_codes import JsOp, JsAttr
from struct import pack

from jaseci.utils.utils import parse_str_token


def byte_length(val):
    if type(val) == str:
        return len(val)
    else:
        return (val.bit_length() + 7) // 8


def to_bytes(val):
    if type(val) == str:
        return bytes(val, "utf-8")
    elif type(val) == float:
        return pack("d", val)
    elif type(val) == int:
        return val.to_bytes(byte_length(val), "little")


class CodeGenPass(IrPass):
    def __init__(self, debug_info=False, **kwargs):
        super().__init__(**kwargs)
        self.debug_info = debug_info
        self.cur_loc = None

    def emit(self, node, *items):
        if not hasattr(node, "bytecode"):
            node.bytecode = bytearray()
        for i in items:
            if type(i) is bytes:
                node.bytecode += bytearray(i)
            elif type(i) is str:
                node.bytecode += bytearray(i, "utf-8")
            else:
                node.bytecode.append(i)

    def emit_debug_info(self, node):
        node_loc = [node.loc[0], node.loc[2]]
        if not self.debug_info or self.cur_loc == node_loc:
            return
        self.emit(
            node,
            JsOp.DEBUG_INFO,
            byte_length(node_loc[0]),
            to_bytes(node_loc[0]),
        )
        if not self.cur_loc or self.cur_loc[1] != node_loc[1]:
            self.emit(node, byte_length(node_loc[1]), to_bytes(node_loc[1]))
        else:
            self.emit(node, 0)
        self.cur_loc = node_loc

    def enter_node(self, node):
        # print("entering", node)
        if hasattr(self, f"enter_{node.name}"):
            self.emit_debug_info(node)
            getattr(self, f"enter_{node.name}")(node)

    def exit_node(self, node):
        # print("exiting", node)
        if hasattr(self, f"exit_{node.name}"):
            self.emit_debug_info(node)
            getattr(self, f"exit_{node.name}")(node)

    def exit_atom(self, node):
        if len(node.kid) == 1 and hasattr(node.kid[0], "bytecode"):
            node.bytecode = node.kid[0].bytecode

    def exit_any_type(self, node):
        node.bytecode = node.kid[0].bytecode

    def exit_TYP_STRING(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.STRING)

    def exit_TYP_INT(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.INT)

    def exit_TYP_FLOAT(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.FLOAT)

    def exit_TYP_LIST(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.LIST)

    def exit_TYP_DICT(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.DICT)

    def exit_TYP_BOOL(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.BOOL)

    def exit_KW_NODE(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.NODE)

    def exit_KW_EDGE(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.EDGE)

    def exit_KW_TYPE(self, node):  # noqa
        self.emit(node, JsOp.LOAD_CONST, JsAttr.TYPE, JsAttr.TYPE)

    def exit_INT(self, node):  # noqa
        val = int(node.token_text())
        self.emit(
            node,
            JsOp.LOAD_CONST,
            JsAttr.INT,
            byte_length(val),
            to_bytes(val),
        )

    def exit_FLOAT(self, node):  # noqa
        val = float(node.token_text())
        self.emit(
            node,
            JsOp.LOAD_CONST,
            JsAttr.FLOAT,
            to_bytes(val),
        )

    def exit_STRING(self, node):  # noqa
        val = parse_str_token(node.token_text())
        self.emit(
            node,
            JsOp.LOAD_CONST,
            JsAttr.STRING,
            byte_length(val),
        )
        if byte_length(val) > 0:
            self.emit(node, to_bytes(val))

    def exit_BOOL(self, node):  # noqa
        val = int(node.token_text() == "true")
        self.emit(
            node,
            JsOp.LOAD_CONST,
            JsAttr.BOOL,
            val,
        )

    def exit_NULL(self, node):  # noqa
        self.emit(
            node,
            JsOp.LOAD_CONST,
            JsAttr.TYPE,
            JsAttr.NULL,
        )



from enum import IntEnum, auto
from jaseci.graph.edge import Edge
from jaseci.graph.node import Node


class JsOp(IntEnum):
    PUSH_SCOPE = auto()
    POP_SCOPE = auto()
    ADD = auto()
    SUB = auto()
    LOAD_CONST = auto()  # [type, bytes, (val)] / [type, (val)] / [type=type, type]
    # LOAD_VAR = auto()  # [bytes, (name)]
    REPORT = auto()  # [type, bytes, (val)]
    ACTION_CALL = auto()
    DEBUG_INFO = auto()  # [bytes, line, bytes, (jacfile)]


class JsAttr(IntEnum):
    TYPE = auto()
    INT = auto()
    FLOAT = auto()
    STRING = auto()
    LIST = auto()
    DICT = auto()
    BOOL = auto()
    NULL = auto()
    NODE = auto()
    EDGE = auto()


type_map = {
    JsAttr.TYPE: type,
    JsAttr.INT: int,
    JsAttr.FLOAT: float,
    JsAttr.STRING: str,
    JsAttr.LIST: list,
    JsAttr.DICT: dict,
    JsAttr.BOOL: bool,
    JsAttr.NULL: None,
    JsAttr.NODE: Node,
    JsAttr.EDGE: Edge,
}



from jaseci.jac.jsci_vm.op_codes import JsOp, JsAttr, type_map
from jaseci.jac.machine.machine_state import MachineState
from jaseci.jac.jsci_vm.inst_ptr import InstPtr, from_bytes
from jaseci.jac.machine.jac_value import JacValue
from jaseci.jac.jsci_vm.disasm import DisAsm


class Stack(object):
    def __init__(self):
        self._stk = []

    def stack_is_empty(self) -> bool:
        return len(self._stk) == 0

    def pop(self):
        if self.stack_is_empty():
            raise Exception("JaseciMachine stack is empty")
        return self._stk.pop()

    def push(self, value):
        self._stk.append(value)

    def peak_stack(self, count=0):
        if not count:
            print(list(reversed(self._stk)))
        else:
            print(list(reversed(self._stk))[0:count])


class VirtualMachine(MachineState, Stack, InstPtr):
    def __init__(self, **kwargs):
        Stack.__init__(self)
        InstPtr.__init__(self)
        MachineState.__init__(self, **kwargs)
        self._op = self.build_op_call()
        self._cur_loc = None

    def reset_vm(self):
        Stack.__init__(self)
        InstPtr.__init__(self)
        self._cur_loc = None
        self._op = self.build_op_call()

    def build_op_call(self):
        op_map = {}
        for op in JsOp:
            op_map[op] = getattr(self, f"op_{op.name}")
        return op_map

    def run_bytecode(self, bytecode):
        self.reset_vm()
        self._bytecode = bytearray(bytecode)
        while self._ip < len(self._bytecode):
            self._op[self._bytecode[self._ip]]()
            self._ip += 1

    def disassemble(self):
        DisAsm().disassemble(self._bytecode)

    def op_PUSH_SCOPE(self):  # noqa
        pass

    def op_POP_SCOPE(self):  # noqa
        pass

    def op_ADD(self):  # noqa
        self.push(self.pop() + self.pop())

    def op_SUB(self):  # noqa
        rhs = self.pop()
        self.push(self.pop() - rhs)

    def op_LOAD_CONST(self):  # noqa
        typ = JsAttr(self.offset(1))
        operand2 = self.offset(2)
        if typ in [JsAttr.TYPE]:
            val = type_map[JsAttr(operand2)]
            self._ip += 2
        elif typ in [JsAttr.INT, JsAttr.STRING]:
            val = from_bytes(type_map[typ], self.offset(3, operand2))
            self._ip += 2 + operand2
        elif typ in [JsAttr.FLOAT]:
            val = from_bytes(float, self.offset(2, 8))
            self._ip += 2 + 8
        elif typ in [JsAttr.BOOL]:
            val = bool(self.offset(2))
            self._ip += 2 + 1
        self.push(JacValue(self, value=val))

    def op_REPORT(self):  # noqa
        self.report.append(self.pop())

    def op_ACTION_CALL(self):  # noqa
        pass

    def op_DEBUG_INFO(self):  # noqa
        byte_len_l = self.offset(1)
        line = from_bytes(int, self.offset(2, byte_len_l))
        byte_len_f = self.offset(3)
        jacfile = from_bytes(str, self.offset(4, byte_len_f)) if byte_len_f else 0
        self._cur_loc = [line, jacfile]
        self._ip += 2 + byte_len_l + byte_len_f



from jaseci.utils.test_core import CoreTest


class TestCodegen(CoreTest):
    """Unit tests for Jac Walker APIs"""

    fixture_src = __file__

    def test_simple_codegen(self):
        self.call(
            self.mast,
            ["sentinel_register", {"code": self.load_jac("simple.jac")}],
        )
        ret = self.call(self.mast, ["walker_run", {"name": "most_basic"}])
        # self.log(ret)
        # from jaseci.jac.ir.passes import PrinterPass
        # PrinterPass(self.mast.active_snt()._jac_ast).run()
        self.assertEqual(ret["report"][0], 5004)
        self.assertEqual(ret["report"][1], "ab")

    # def test_pack_it(self):
    #     self.logger_on()
    #     ret = self.call(
    #         self.mast,
    #         ["sentinel_register", {"code": self.load_jac("simple.jac")}],
    #     )
    #     self.log(ret)
    #     ret = self.call(self.mast, ["walker_run", {"name": "pack_it"}])

    #     from jaseci.jac.ir.passes import PrinterPass

    #     PrinterPass(self.mast.active_snt()._jac_ast).run()
    #     # self.assertEqual(ret["report"][0], 5004)
    #     # self.assertEqual(ret["report"][1], "ab")
